<!DOCTYPE html>
<html>

<head>
  <title>Cosmic-CoNN CR Detector - Web App</title>
  <script src="https://code.jquery.com/jquery-3.5.1.min.js"
    integrity="sha256-9/aliU8dGd2tb6OSsuzixeV4y/faTqgFtohetphbbj0=" crossorigin="anonymous"></script>
  <script src="https://cdnjs.cloudflare.com/ajax/libs/popper.js/1.14.3/umd/popper.min.js"
    integrity="sha384-ZMP7rVo3mIykV+2+9J3UJ46jBk0WLaUAdn689aCwoqbBJiSnjAK/l8WvCWPIPm49"
    crossorigin="anonymous"></script>
  <script src="https://stackpath.bootstrapcdn.com/bootstrap/4.1.3/js/bootstrap.min.js"
    integrity="sha384-ChfqqxuZUCnJSK3+MXmPNIyE6ZbWh2IMqE241rYiqJxyMiZ6OW/JmZQ5stwEULTy"
    crossorigin="anonymous"></script>
  <link rel="stylesheet" href="https://stackpath.bootstrapcdn.com/bootstrap/4.1.3/css/bootstrap.min.css"
    integrity="sha384-MCw98/SFnGE8fJT3GXwEOngsV7Zt27NXFoaoApmYm81iuXoPkFOJwJ8ERdknLPMO" crossorigin="anonymous">
  <link rel="stylesheet" href="{{url_for('static', filename='css/controls.css')}}">
  <link rel="stylesheet" href="{{url_for('static', filename='css/panel.css')}}">
  <link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/@splidejs/splide@latest/dist/css/splide.min.css">
  <link rel="stylesheet" href="{{url_for('static', filename='css/splide-skyblue.min.css')}}">
  <link rel="stylesheet" href="{{url_for('static', filename='css/thumbnails.css')}}">

  <!-- Global site tag (gtag.js) - Google Analytics -->
  <script async src="https://www.googletagmanager.com/gtag/js?id=G-5VZNWK8ZPN"></script>
  <script>
    window.dataLayer = window.dataLayer || [];
    function gtag() { dataLayer.push(arguments); }
    gtag('js', new Date());

    gtag('config', 'G-5VZNWK8ZPN');
  </script>

</head>

<style>
  ::-webkit-scrollbar {
    -webkit-appearance: none;
    width: 14px;
  }

  ::-webkit-scrollbar-thumb {
    border-radius: 4px;
    background-color: rgba(0,0,0,.5);
    -webkit-box-shadow: 0 0 1px rgba(255,255,255,.5);
}
</style>

<body>
  <div class="container" style="max-width: 1280px;">
    <div class="row col-12 justify-content-center mx-auto" style="">

      {% include 'components/HeaderContainer.html'%}
      {% include 'components/ThumbnailContainer.html'%}
      {% include 'components/PreviewContainer.html'%}

    </div>
    <script src="https://cdn.jsdelivr.net/npm/@splidejs/splide@latest/dist/js/splide.min.js"></script>
    <script src="https://cdn.jsdelivr.net/gh/nicolaspanel/numjs@0.15.1/dist/numjs.min.js"></script>
    <script async src="https://docs.opencv.org/master/opencv.js" type="text/javascript"></script>
    <script src="{{url_for('static', filename='GlobalParas.js')}}"></script>
    <script src="{{url_for('static', filename='ThumbnailBar.js')}}"></script>
    <script src="{{url_for('static', filename='Utils.js')}}"></script>
    <script src="{{url_for('static', filename='ScrollingControl.js')}}"></script>
    <script src="{{url_for('static', filename='ImageUtils.js')}}"></script>
    <script src="{{url_for('static', filename='ControlPanel.js')}}"></script>
    <script src="{{url_for('static', filename='AstroMessage.js')}}"></script>
    <script src="{{url_for('static', filename='StatusIndicator.js')}}"></script>
    <script src="{{url_for('static', filename='ImageModel.js')}}"></script>
    <script src="{{url_for('static', filename='ImageView.js')}}"></script>
    <script src="{{url_for('static', filename='ImageController.js')}}"></script>
    <script src="{{url_for('static', filename='FileInputHandler.js')}}"></script>
    <script src="{{url_for('static', filename='DownloadHandler.js')}}"></script>
    <script src="{{url_for('static', filename='script.js')}}"></script>
</body>

</html>