.overview-image {
    border-radius: 0.2rem;
    box-shadow: 1px 1px 5px -1px rgba(0, 0, 0, 0.35);
}

.zoom-image {
    border-radius: 0.25rem;
    box-shadow: 1px 2px 7px -1px rgba(0, 0, 0, 0.35);
}

.status-badge {
    box-shadow: 1px 2px 7px -1px rgba(0, 0, 0, 0.25);
}

.pixel-value-indicator {
    font-size: 0.75rem;
    color: rgb(103, 255, 179);
}