<!-- Dilation Area -->
<div id="dilation-control-area" class="control-group btn-group d-flex mr-2 control-shadow">
    <span id="dilation-label" class=" control-label rounded-left">Dilation</span>
    <input id="dilation-value" class="d-inline-block text-center control-input" placeholder="" type="text">
    <button id="dilation-plus-button" class="btn btn-primary btn-sm control-normal-button">+</button>
    <button id="dilation-minus-button" class="btn btn-primary btn-sm control-normal-button">-</button>
</div>

<!-- Threashold -->
<div id="dilation-control-area" class="control-group btn-group d-flex mr-2 control-shadow">
    <span id="threshold-label" class=" control-label rounded-left">Threshold</span>
    <input id="threshold-value" class="d-inline-block text-center control-input" placeholder="" type="text">
    <button id="threshold-plus-button" class="btn btn-primary btn-sm control-normal-button">+</button>
    <button id="threshold-minus-button" class="btn btn-primary btn-sm control-normal-button">-</button>
</div>