"""Top-level package for cosmic_conn."""

__author__ = """Chengyuan Xu, Curtis McCully, Boning Dong, D. Andrew Howell, and Pradeep Sen"""
__email__ = "cxu@ucsb.edu"
__version__ = "0.2.5"

from cosmic_conn.inference_cr import init_model, detect_image, detect_FITS