<div class="row col-12 px-0 align-items-center my-2">

  <template id="thumbnail-element-template">
    {% include 'components/Thumbnail.html'%}
  </template>
  <h1 class="ml-4" style="color: rgb(78, 78, 78); font-size: 16px;">CR Thumbnails</h1>
  <div id="thumbnail-bar" class="splide px-4" style="width: 100%;">
    <div class="splide__track">
      <ul class="splide__list" id="thumbnail-container">
        {% include 'components/Thumbnail.html'%}
        {% include 'components/Thumbnail.html'%}
        {% include 'components/Thumbnail.html'%}
        {% include 'components/Thumbnail.html'%}
        {% include 'components/Thumbnail.html'%}
        {% include 'components/Thumbnail.html'%}
        {% include 'components/Thumbnail.html'%}
        {% include 'components/Thumbnail.html'%}
        {% include 'components/Thumbnail.html'%}
        {% include 'components/Thumbnail.html'%}
        {% include 'components/Thumbnail.html'%}
        {% include 'components/Thumbnail.html'%}
      </ul>
    </div>
  </div>
</div>