<li class="splide__slide">
    <div>
        <img class="thumbnail" src="https://via.placeholder.com/32x32?text=."
            style="width: 100%; height: auto; image-rendering: pixelated;">
    </div>
</li>